using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.Graphics.OpenGL;
using SlatedGameToolkit.Framework.Graphics.Window;

namespace SlatedGameToolkit.Framework.AssetSystem {
    /// <summary>
    /// A delegate that should modify a given path.
    /// </summary>
    /// <param name="path">A path to modify.</param>
    /// <returns>The modified path.</returns>
    public delegate string ModifyPath(string path);

    /// <summary>
    /// A delegate that performs the loading of the specific asset.
    /// </summary>
    /// <param name="name">The path of the asset.</param>
    /// <typeparam name="T">The type of the resulting asset.</typeparam>
    /// <returns>A loaded, useable asset.</returns>
    public delegate IAssetData LoadAsset(string path);

    //TODO Implement change to two step asset pipeline.

    public class AssetManager {
        public IAssetUseable this[string key] {
            get {
                if (!this.assets.ContainsKey(key)) { // if the data is still uninitialized, then do so.
                    InitializeData(key);
                }
                return this.assets[key];
            }
        }
        private readonly object assetThreadLock = new object();
        private Thread thread;
        private ConcurrentQueue<string> loadables;
        private volatile bool load;
        public bool Complete {
            get {
                return loadables.Count == 0;
            }
        }
        private ConcurrentDictionary<string, IAssetUseable> assets;
        
        /// <summary>
        /// A concurrent dictionary containing all the path modifier's in association with their extension.
        /// This is what the manager uses to determine how to modify a filename or path dependent on it's extension.
        /// </summary>
        /// <value>A dictionary of file extension associated with respective paths.</value>
        public ConcurrentDictionary<string, ModifyPath> PathModifiers { get; private set; }

        /// <summary>
        /// If this value is not null, the asset manager will modify paths according to this delegate if their is not a more appropriate modifier.
        /// </summary>
        /// <value>The delegate to use to modify the path.</value>
        public ModifyPath DefaultPathModifier {get; set;}

        /// <summary>
        /// A dictionary containing associations between file extensions and their loaders.
        /// All file extensions will be requested in lower cases, and therefore, the extension key in this dictionary should also be all lowercase.
        /// </summary>
        /// <value>The dictionary that associates the loaders to their file extensions.</value>
        public ConcurrentDictionary<string, LoadAsset> Loaders { get ; private set; }
        private ConcurrentDictionary<string, IAssetData> uninitializedDataDictionary;
        private ConcurrentBag<string> uninitializedDataBag;

        private GLContext glContext;
        public AssetManager(GLContext glContext = null) {
            this.glContext = glContext ?? WindowContextsManager.CurrentGL;
            this.loadables = new ConcurrentQueue<string>();
            this.assets = new ConcurrentDictionary<string, IAssetUseable>();
            this.PathModifiers = new ConcurrentDictionary<string, ModifyPath>();
            this.Loaders = new ConcurrentDictionary<string, LoadAsset>();
            this.uninitializedDataDictionary = new ConcurrentDictionary<string, IAssetData>();
            this.uninitializedDataBag = new ConcurrentBag<string>();

            thread = new Thread(Run);
            this.load = true;
            thread.Name = "Asset-Loader";
            thread.IsBackground = true;
            thread.Start();

        }

        /// <summary>
        /// Queue's the loadable for batched loading.
        /// The file name of the loadable is the one the loadable is saved under.
        /// </summary>
        /// <param name="name">path of file.</param>
        public void QueueLoad(string name) {
            loadables.Enqueue(name);
        }

        private void Run() {
            while (load) {
                lock (assetThreadLock)
                {
                    string loadable;
                    while (loadables.TryDequeue(out loadable)) {
                        Load(loadable);
                    }
                    if (load) Monitor.Wait(assetThreadLock);
                }
            }
        }

        /// <summary>
        /// Loads the loadable and stores it under the filename given to the loadable.
        /// </summary>
        /// <param name="name">The loadable to load.</param>
        public void Load(string name) {
            if (glContext == null) glContext = WindowContextsManager.CurrentGL;
            string original = name;
            string ext = Path.GetExtension(name).ToLower().Substring(1);
            if (PathModifiers.ContainsKey(ext)) {
                name = PathModifiers[ext](name);
            } else if (DefaultPathModifier != null) {
                name = DefaultPathModifier(name);
            }
            if (!Loaders.ContainsKey(ext)) throw new FrameworkUsageException(string.Format("Failed to find associated loader for file \"{0}\" with extension \"{1}\".", name, ext));
            uninitializedDataDictionary[original] = Loaders[ext](name);
            uninitializedDataBag.Add(name);
            GameEngine.ToRunQueue.Enqueue(InitializeDataRandom);
        }

        private void InitializeDataRandom() {
            string name;
            if (uninitializedDataBag.TryTake(out name)) {
                if (this.uninitializedDataDictionary.ContainsKey(name)) {
                    this.assets[name] = this.uninitializedDataDictionary[name].InitializeUseable(glContext);
                }
            } else {
                throw new InternalFrameworkException("Difference in queued uninitialized data loading quantity and actual quantity.");
            }
        }

        private void InitializeData(string name) {
            this.assets[name] = this.uninitializedDataDictionary[name].InitializeUseable(glContext);
        }

        /// <summary>
        /// Loads the currently queued asynchronously.
        /// </summary>
        public void LoadQueued() {
            if (Monitor.TryEnter(assetThreadLock)) {
                Monitor.Pulse(assetThreadLock);
            }
        }

        public void WaitAndLoadAllQueued() {
            LoadQueued();
            Monitor.Enter(assetThreadLock);
            Monitor.Exit(assetThreadLock);
            return;
        }

        /// <summary>
        /// Unloads the asset under the given name.
        /// This disposes the asset as well as removes it from the manager.
        /// </summary>
        /// <param name="name">The name of the asset to unload.</param>
        public void Unload(string name) {
            IAssetUseable asset;
            assets.TryRemove(name, out asset);
            asset.Dispose();
        }

        /// <summary>
        /// Removes all the assets.
        /// </summary>
        public void UnloadAll() {
            foreach (string name in assets.Keys)
            {
                Unload(name);
            }
        }
    }
}