using System;
using System.Drawing;
using System.Numerics;
using SlatedGameToolkit.Framework.Graphics.Textures;

namespace SlatedGameToolkit.Framework.Graphics.Render
{
    public struct RectangleMesh : IMesh
    {
        private Matrix4x4 matRot;
        private bool changed;
        private RectangleF rectangle;
        private Vector3 rotation;
        private RectangleF textureBounds;
        private ValueTuple<Vector3, Vector2>[] vertices;
        private uint[] indices;

        public ValueTuple<Vector3, Vector2>[] Vertices
        {
            get {
                if (changed) CalculateVertices();
                return vertices;
            }
        }

        public float X {
            get 
            {
                return rectangle.X;
            }

            set 
            {
                changed = true;
                rectangle.X = value;
            }
        }

        public float Y {
            get
            {
                return rectangle.Y;
            }

            set
            {
                changed = true;
                rectangle.Y = value;
            }
        }

        public float Width
        {
            get
            {
                return rectangle.Width;
            }

            set
            {
                changed = true;
                rectangle.Width = value;
            }
        }

        public float Height
        {
            get
            {
                return rectangle.Height;
            }

            
            set
            {
                changed = true;
                rectangle.Height = value;
            }
        }

        public RectangleF Bounds {
            get {
                return new RectangleF(X, Y, Width, Height);
            }

            set {
                X = value.X;
                Y = value.Y;
                Width = value.Width;
                Height = value.Height;
            }
        }

        public RectangleF TextureBounds
        {
            get
            {
                return textureBounds;
            }
            set
            {
                this.vertices[0].Item2.X = value.X;
                this.vertices[0].Item2.Y = value.Y;

                this.vertices[1].Item2.X = value.X + value.Width;
                this.vertices[1].Item2.Y = value.Y;
                
                this.vertices[2].Item2.X = value.X + value.Width;
                this.vertices[2].Item2.Y = value.Y + value.Height;
                
                this.vertices[3].Item2.X = value.X;
                this.vertices[3].Item2.Y = value.Y + value.Height;

                this.textureBounds = value;
            }
        }

        public uint[] Elements { get {return indices; } }

        public Vector3 Rotation 
        {
            get 
            {
                return rotation;
            }

            set
            {
                changed = true;
                rotation = value;
                matRot = Matrix4x4.CreateFromYawPitchRoll(value.X, value.Y, value.Z);
            }
        }

        public ITexture Texture { get; set; }

        public Color Color { get; set; }


        public RectangleMesh(RectangleF meshBounds, ITexture texture, Color color) {
            this.changed = true;
            this.rotation = Vector3.Zero;
            this.rectangle = RectangleF.Empty;
            this.Texture = texture;
            this.Color = color;
            this.indices = new uint[] {0, 1, 3, 1, 2, 3};
            vertices = new ValueTuple<Vector3, Vector2>[4];
            this.textureBounds = new RectangleF(0, 0, 1, 1);
            this.matRot = Matrix4x4.Identity;

            this.TextureBounds = textureBounds;
            this.Bounds = meshBounds;
        }

        public RectangleMesh(ITexture texture, Color color) : this(new RectangleF(0,0,0,0), texture, color)
        {
            
        }

        public RectangleMesh(RectangleF meshBounds, RectangleF textureRegion, ITexture texture, Color color) : this(meshBounds, texture, color) {
            this.TextureBounds = textureRegion;
        }
        
        
        private void CalculateVertices() {
            if (!changed) return;
            Vector3[] baseVerts = new Vector3[4];
            baseVerts[0] = new Vector3(this.rectangle.X, this.rectangle.Y, 0);
            baseVerts[1] = new Vector3(this.rectangle.Right, this.rectangle.Y, 0);
            baseVerts[2] = new Vector3(baseVerts[1].X, this.rectangle.Bottom, 0);
            baseVerts[3] = new Vector3(this.rectangle.X, baseVerts[2].Y, 0);

            for (int i = 0; i < vertices.Length; i++)
            {
                vertices[i].Item1 = Vector3.Transform(baseVerts[i], matRot);
            }
            changed = false;
        }
    }
}