using SlatedGameToolkit.Framework;
using SlatedGameToolkit.Framework.Logging;
using SlatedGameToolkit.Framework.StateSystem;
using SlatedGameToolkit.Tools.CommandSystem;
using SlatedGameToolkit.Tools.CommandSystem.Interaction;
using SlatedGameToolkit.Tools.Utilities.Playground;

namespace SlatedGameToolkit.Tools.Commands
{
    public class GraphicalPlaygroundCommand : IInvocable
    {
        private readonly string[] invokers = new string[] {"playground"};
        private ConsolePlaygroundListener logListener;

        public bool Execute(IInteractable interactable, string[] args)
        {
            if (args.Length < 1) return false;
            args[0] = args[0].ToLower();
            if (args[0].Equals("start")) {
                if (GameEngine.IsRunning()) {
                    interactable.Tell("Engine is already running!");
                    return true;
                }
                GameEngine.Ignite(new MainState());
                return true;
            } else if (args[0].Equals("stop")) {
                if (!GameEngine.IsRunning()) {
                    interactable.Tell("Engine was never running!");
                    return true;
                }
                GameEngine.Stop();
                return true;
            } else if (args[0].Equals("status")) {
                interactable.Tell("Running: " + GameEngine.IsRunning());
                interactable.Tell("Target FPS: " + (GameEngine.targetFPS <= 0 ? "Not bounded." : GameEngine.targetFPS.ToString()));
                interactable.Tell("Update Step: " + (GameEngine.UpdatesPerSecond <= 0 ? "Not Locked." : GameEngine.UpdatesPerSecond.ToString()));
                interactable.Tell("Debug logging: " + ((logListener == null) ? "false" : "true"));
                return true;
            } else if (args[0].Equals("debug")) {
                if (logListener == null) {
                    Logger.AddLogListener((logListener = new ConsolePlaygroundListener(interactable as ConsoleInteraction)));
                    logListener.Debug = true;
                    interactable.Tell("Attached log listener.");
                } else {
                    logListener.Debug = !logListener.Debug;
                }
                interactable.Tell(string.Format("Debug logging has been turned {0}.", logListener.Debug ? "on" : "off"));
                return true;
            } else if (args[0].Equals("log")) {
                if (logListener == null) {
                    Logger.AddLogListener((logListener = new ConsolePlaygroundListener(interactable as ConsoleInteraction)));
                    interactable.Tell("Listening to game engine's logging.");
                } else {
                    Logger.RemoveLogListener(logListener);
                    logListener = null;
                    interactable.Tell("Stopped listening to game engine's logging.");
                }
                return true;
            } else if (args[0].Equals("updates")) {
                int updates;
                if (args.Length < 2 || !int.TryParse(args[1], out updates)) {
                    interactable.Tell("Missing numerical value dictating updates per second.");
                } else {
                    GameEngine.UpdatesPerSecond = updates;
                    interactable.Tell("Target updates per second set to: " + updates);
                }
                return true;
            } else if (args[0].Equals("framerate")) {
                int frames;
                if (args.Length < 2 || !int.TryParse(args[1], out frames)) {
                    interactable.Tell("Missing numerical value dictating framerate.");
                } else {
                    GameEngine.targetFPS = frames;
                    interactable.Tell("Target framerate set to: " + frames);
                }
                return true;
            }
            return false;
        }

        public string getDescription()
        {
            return "Controls the playground engine status.";
        }

        public string[] GetInvokers()
        {
            return invokers;
        }

        public string getUsage(string arg)
        {
            if (arg == null) {
                return "\"playground [start | stop | status | log | debug | framerate <framerate> | updates <update rate>]\"";
            } else if (arg.Equals("start")) {
                return "Starts the playground.";
            } else if (arg.Equals("stop")) {
                return "Stops the playground.";
            } else if (arg.Equals("status")) {
                return "Retrieves the FPS, update rate, whether or not the playground is running, and if debugging is turned on.";
            } else if (arg.Equals("log")) {
                return "Toggles whether or not to attach logging to the console.";
            } else if (arg.Equals("debug")) {
                return "Toggles debugging. If logging is disabled, logging will be enabled.";
            } else if (arg.Equals("framerate")) {
                return "sets the FPS. Needs to be followed by a numerical value representing the new target FPS.";
            } else if (arg.Equals("updates")) {
                return "Sets the new update rate. Needs to be followed by a numerical value representing the target update rate.";
            } else {
                return null;
            }
        }

        public void Dispose()
        {
            GameEngine.Stop();
        }
    }
}